/*
** freebsd.c	   Low level kernel access functions for FreeBSD 2.x
**
** This program is in the public domain and may be used freely by anyone
** who wants to.
**
** Please send bug fixes/bug reports to: Peter Eriksson <pen@lysator.liu.se>
*/

#include "config.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <sys/socketvar.h>

#include <sys/queue.h>
#include <sys/uio.h>

#include <paths.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <nlist.h>
#include <pwd.h>
#include <signal.h>
#include <syslog.h>

#include <kvm.h>

#define KERNEL

#include <sys/file.h>

#undef KERNEL

#include <fcntl.h>

#include <sys/time.h>
#include <sys/user.h>
#include <sys/wait.h>

#include <sys/filedesc.h>
#include <sys/proc.h>

#include <net/if.h>
#include <net/route.h>
#include <netinet/in.h>

#include <netinet/in_systm.h>
#include <netinet/ip.h>
#if defined(HAVE_IPV6) && __FreeBSD__ < 3
#include <netinet6/ip6.h>
#endif

#include <netinet/in_pcb.h>
#if defined(HAVE_IPV6) && __FreeBSD__ < 3
#include <netinet6/in6_pcb.h>
#endif

#include <netinet/tcp.h>
#include <netinet/ip_var.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>

#include <arpa/inet.h>

#include <pwd.h>

#include "pidentd.h"


struct kainfo {
	kvm_t *kd;
	int nfile;
	struct nlist nl[3];
};

static int getbuf(struct kainfo *, u_long, char *, u_int, char *);

#ifdef HAVE_IPV6

#if __FreeBSD__ < 3
static struct socket *getlist(struct kainfo *, struct in6pcb *,
    struct in6_addr *, int, struct in6_addr *, int);

static struct in6pcb tcb;
#else
static struct socket *getlist(struct kainfo *, struct inpcbhead *,
    struct in6_addr *, int, struct in6_addr *, int);

static struct inpcbhead tcb;
#endif

#else

static struct socket *getlist(struct kainfo *, struct inpcbhead *,
    struct in_addr *, int, struct in_addr *, int);

static struct inpcbhead tcb;

#endif

int
ka_init(void)
{
	/* The kvm routines are not MT-Safe! */
	kernel_threads = 1;
	return 0;
}

int
ka_open(misc)
	void **misc;
{
	struct kainfo *kp;


	A_NEW(kp);

	/*
	** Open the kernel memory device
	*/
	if ((kp->kd = (kvm_t *)kvm_openfiles(NULL, NULL,
	    NULL, O_RDONLY, NULL)) == NULL) {
		syslog(LOG_ERR, "kvm_open: %m");
		a_free(kp);
		return -1;
	}

#define N_TCB 0
#define N_BTEXT 1

#if defined(HAVE_IPV6) && __FreeBSD__ < 3
	kp->nl[N_TCB].n_name = "_tcb6";
#else
	kp->nl[N_TCB].n_name = "_tcb";
#endif

	kp->nl[N_BTEXT].n_name = "_btext";
	kp->nl[2].n_name = "";
	/*
	** Extract offsets to the needed variables in the kernel
	*/
	if (kvm_nlist(kp->kd, kp->nl) < 0) {
		syslog(LOG_ERR, "kvm_nlist: %m");
		kvm_close(kp->kd);
		a_free(kp);
		return -1;
	}

	*misc = (void *) kp;
	return 0;
}


/*
** Get a piece of kernel memory with error handling.
** Returns 1 if call succeeded, else 0 (zero).
*/
static int
getbuf(kp, addr, buf, len, what)
	struct kainfo *kp;
	unsigned long addr;
	char *buf;
	unsigned int len;
	char *what;
{
	if (addr < kp->nl[N_BTEXT].n_value ||		/* Overkill.. */
	    addr >= (unsigned long)0xFFC00000 ||
	    (addr + len) < kp->nl[N_BTEXT].n_value ||
	    (addr + len) >= (unsigned long)0xFFC00000) {
	    	syslog(LOG_ERR,
		    "getbuf: bad address (%08x not in %08x-0xFFC00000) - %s",
		addr, kp->nl[N_BTEXT].n_value, what);
		return 0;
	}
    
	if (kvm_read(kp->kd, addr, buf, len) < 0) {
		syslog(LOG_ERR, "getbuf: kvm_read(%08x, %d) - %s : %m",
		    addr, len, what);
	
		return 0;
	}
    
	return 1;
}

/*
** Traverse the inpcb list until a match is found.
** Returns NULL if no match.
*/
static struct socket *
#if defined(HAVE_IPV6) && __FreeBSD__ < 3
getlist(kp, pcbp, faddr, fport, laddr, lport)
	struct kainfo *kp;
	struct in6pcb *pcbp;
#else
getlist(kp, pcbhead, faddr, fport, laddr, lport)
	struct kainfo *kp;
	struct inpcbhead *pcbhead;
#endif
#ifdef HAVE_IPV6
	struct in6_addr *faddr;
	int fport;
	struct in6_addr *laddr;
	int lport;
#else
	struct in_addr *faddr;
	int fport;
	struct in_addr *laddr;
	int lport;
#endif
{
#if defined(HAVE_IPV6) && __FreeBSD__ < 3
	struct in6pcb *head;

	if (!pcbp)
		return NULL;

	head = pcbp->in6p_prev;
	do {
		if (!memcmp(pcbp->in6p_faddr.s6_addr, faddr->s6_addr,
			    sizeof(struct in6_addr)) &&
		    !memcmp(pcbp->in6p_laddr.s6_addr, laddr->s6_addr,
			    sizeof(struct in6_addr)) &&
		     pcbp->in6p_fport         == fport &&
		     pcbp->in6p_lport         == lport)
			return pcbp->in6p_socket;
	} while (pcbp->in6p_next != head &&
		 getbuf(kp, (long)pcbp->in6p_next,
			(char *)pcbp, sizeof(struct in6pcb), "tcblist"));
#else
#ifdef HAVE_IPV6
	struct in6pcb *head, pcbp;
#else
	struct inpcb *head, pcbp;
#endif

	head = pcbhead->lh_first;
	if (!head)
		return NULL;

	do {
		if (!getbuf(kp, (u_long)head, (char *)&pcbp,
		    sizeof(struct inpcb), "tcblist"))
			break;
#ifdef HAVE_IPV6
		if (pcbp.inp_vflag == INP_IPV4) {
			if (IN6_IS_ADDR_V4MAPPED(faddr) &&
			    pcbp.inp_faddr.s_addr == faddr->s6_addr32[3] &&
			    pcbp.inp_fport        == fport &&
			    pcbp.inp_lport        == lport )
				return(pcbp.inp_socket);
		} else if (pcbp.inp_vflag == INP_IPV6 &&
			   !memcmp(pcbp.in6p_faddr.s6_addr, faddr->s6_addr,
				   sizeof(struct in6_addr)) &&
#else
		if (pcbp.inp_faddr.s_addr == faddr->s_addr &&
#endif
		    pcbp.inp_fport        == fport &&
		    pcbp.inp_lport        == lport )
			return(pcbp.inp_socket);
		head = pcbp.inp_list.le_next;
	} while (head != NULL);
#endif

	return NULL;
}

/*
** Return the user number for the connection owner
*/
int
ka_lookup(vp, ke)
	void *vp;
	struct kernel *ke;
{
	struct kainfo *kp = (struct kainfo *) vp;
	struct socket *sockp;
	struct kinfo_proc	*kgp;
	int i, nentries;
	
#if 0
	int fport, lport;

#ifdef HAVE_IPV6
	struct in6_addr *faddr6, *laddr6;
#endif
	struct in_addr *faddr, *laddr;

	fport = SGPORT(ke->remote);
	lport = SGPORT(ke->local);

	if (SGFAM(ke->remote) == AF_INET)
	{
	    faddr = SGADDRP(ke->remote);
	    laddr = SGADDRP(ke->local);
	}
	else
	{
	    faddr6 = SGADDRP(ke->remote);
	    laddr6 = SGADDRP(ke->local);
	}
#endif
	if ((kgp = kvm_getprocs(kp->kd, KERN_PROC_ALL, 0, &nentries)) == NULL)
	{
		syslog(LOG_ERR, "kvm_getprocs: %m");
		return -1;
	}

	/* -------------------- TCP PCB LIST -------------------- */
	if (!getbuf(kp, kp->nl[N_TCB].n_value, (char*)&tcb, sizeof(tcb), "tcb"))
		return -1;

#if defined(HAVE_IPV6) && __FreeBSD__ < 3
	tcb.in6p_prev = (struct in6pcb *)kp->nl[N_TCB].n_value;
#endif
	sockp = getlist(kp, &tcb,
			SGADDRP(ke->remote),
			SGPORT(ke->remote),
			SGADDRP(ke->local),
			SGPORT(ke->local));

	if (!sockp)
		return -1;

	/*
	** Locate the file descriptor that has the socket in question
	** open so that we can get the 'ucred' information
	*/
	for (i = 0; i < nentries; i++) {
		if(kgp[i].kp_proc.p_fd != NULL) {
			struct filedesc	pfd;
			struct file	**ofiles;
			struct file	ofile;
			int j;

			if(!getbuf(kp, (u_long)kgp[i].kp_proc.p_fd,
			    (char *)&pfd, sizeof(pfd), "pfd"))
			        return(-1);

			ofiles = (struct file **) A_ALLOC(pfd.fd_nfiles *
			    sizeof(struct file *));
      
			if(!getbuf(kp, (u_long)pfd.fd_ofiles, (char *)ofiles,
			    pfd.fd_nfiles * sizeof(struct file *), "ofiles")) {
				a_free(ofiles);
				return(-1);
			}

			for(j = 0; j < pfd.fd_nfiles; j ++) {
				if (!ofiles[j])	/* might be sparse */
					continue;
	  
				if(!getbuf(kp, (u_long)ofiles[j], (char *)&ofile,
				    sizeof(struct file), "ofile")) {
					a_free(ofiles);
					return(-1);
				}

				if (ofile.f_count == 0)
					continue;

				if (ofile.f_type == DTYPE_SOCKET &&
					(struct socket *) ofile.f_data == sockp) {

					ke->ruid = kgp[i].kp_eproc.e_pcred.p_ruid;
					ke->euid = kgp[i].kp_eproc.e_ucred.cr_uid;
					a_free(ofiles);
					return 0;
				}
			}

			a_free(ofiles);
		}
	}

	return -1;
}
