/*
** kernel.c - The kernel access threads
**
** Copyright (c) 1997-2002 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#include <stdio.h>
#include <errno.h>
#include <syslog.h>
#include <string.h>
#include <stdlib.h>

#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "pidentd.h"


int kernel_threads = 4;
int kernel_buffers = 16;
int kernel_attempts = 5;


static PQUEUE kbuf_request;
static PQUEUE kbuf_free;


#ifndef HAVE_THREADS
static void *ka_data = NULL;
#endif


struct kernel *
kernel_alloc(void)
{
#ifdef HAVE_THREADS
    struct kernel *kp;

    if (pqueue_get(&kbuf_free, (void *) &kp) != 1)
	return NULL;
    
    avail_init(&kp->av, 1);
    
    kp->status = -1;
    kp->euid = NO_UID;
    kp->ruid = NO_UID;

    kp->pid = NO_PID;
    kp->cmd = NULL;
    kp->argv = NULL;
    
    return kp;
#else
    static struct kernel kpb;
    
    kpb.status = -1;
    kpb.euid = NO_UID;
    kpb.ruid = NO_UID;

    kpb.pid = NO_PID;
    kpb.cmd = NULL;
    kpb.argv = NULL;
    
    return &kpb;
#endif
}


void
kernel_free(struct kernel *kp)
{
    a_free(kp->cmd);
    a_free(kp->argv);

#ifdef HAVE_THREADS
    if (pqueue_put(&kbuf_free, kp) != 1)
	syslog(LOG_ERR, "kernel_free: pqueue_put(kbuf_free) failed\n");
#endif
}

#ifdef HAVE_THREADS
static void *
kernel_thread(void *vp)
{
    struct kernel *kp;
    int attempt;
    char nab1[64];

    
    if (debug)
	fprintf(stderr, "kernel_thread() started\n");

    while (pqueue_get(&kbuf_request, (void *) &kp) == 1)
    {
	if (debug > 1)
	{
	    fprintf(stderr, "remote = %s\n",
		    s_inet_ntox(&kp->remote, nab1, sizeof(nab1)));
	    fprintf(stderr, "local = %s\n",
		    s_inet_ntox(&kp->local, nab1, sizeof(nab1)));
	}

	attempt = 0;
	do
	{
	    kp->status = ka_lookup(vp, kp);
	} while (kp->status < 0 && ++attempt < kernel_attempts);
	
	avail_signal(&kp->av);
    }
    
    if (debug)
	fprintf(stderr, "kernel_thread() terminating\n");

    return NULL;
}
#endif


int
kernel_init(void)
{
#ifdef HAVE_THREADS
    int i;
    struct kernel *kp;
    pthread_t tid;
    
    /*
    ** Create the request queue
    */
    if (pqueue_init(&kbuf_request, kernel_buffers) < 0)
    {
	syslog(LOG_ERR, "pqueue_create(%d) failed: %m", kernel_buffers);
	return -1;
    }


    /*
    ** Create the free pool of buffers
    */
    if (pqueue_init(&kbuf_free, kernel_buffers * 2) < 0)
    {
	syslog(LOG_ERR, "pqueue_create(%d) failed: %m", kernel_buffers * 2);
	return -1;
    }

    for (i = 0; i < kernel_buffers * 2; i++)
    {
	kp = a_malloc(sizeof(*kp), "struct kernel");
	
	if (pqueue_put(&kbuf_free, kp) != 1)
	{
	    syslog(LOG_ERR, "kernel_init: pqueue_put() failed");
	    return -1;
	}
    }


    /*
    ** Create the kernel accessing thread workpool
    */
    for (i = 0; i < kernel_threads; i++)
    {
	void *key = NULL;

	if (ka_open(&key) < 0)
	    return -1;
	
	pthread_create(&tid, NULL, &kernel_thread, key);
    }
    
    return 0;
#else
    return ka_open(&ka_data);
#endif
}


void
kernel_query(struct kernel *kp)
{
#ifdef HAVE_THREADS
    if (pqueue_put(&kbuf_request, kp) != 1)
    {
	syslog(LOG_ERR, "kernel_query: pqueue_put() failed");
	return;
    }

    avail_wait(&kp->av);
#else
    int attempt = 0;
    
    if (debug > 2)
	fprintf(stderr, "kernel_query (no threads version): started\n");
    
    do
    {
	kp->status = ka_lookup(ka_data, kp);
    } while (kp->status < 0 && ++attempt < kernel_attempts);
#endif
}


