/*
** request.c - Handle an IDENT protocol request
**
** Copyright (c) 1997-1999 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <syslog.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifndef HAVE_THREADS
#include <sys/types.h>
#include <sys/wait.h>
#endif

#include "pidentd.h"

int extensions_enabled = 0;
int multiquery_enabled = 0;
int reject_pidentd_port = 0;
int allow_proxy_query = 0; /* 0 = not, 1 = via loopback, 2 = everybody */
int request_timeout = 120;  /* 2 minutes */


static int 
handle_one_request(CLIENT *cp,
		   char *buf,
		   FDBUF *fb)
{
    struct kernel *kp;
    char nab[64];
    char *cmd, *bp;
    int fd = cp->fd;
    struct sockaddr_gen *remote_addr = &cp->rsin;
    struct sockaddr_gen *local_addr = &cp->lsin;
    unsigned int local_port, remote_port;
    unsigned int la1, la2, la3, la4;
    unsigned int ra1, ra2, ra3, ra4;
    int proxy_query = 0;
    
    
    if (debug)
	fprintf(stderr, "request_handler: fd#%d: '%s'\n", fd, buf);

    if (sscanf(buf, " %d , %d", &local_port, &remote_port) == 2 ||
	(allow_proxy_query &&
	(proxy_query =
	 (sscanf(buf, " %u.%u.%u.%u:%u , %u.%u.%u.%u:%u",
		 &la1, &la2, &la3, &la4, &local_port,
		 &ra1, &ra2, &ra3, &ra4, &remote_port) == 10)) != 0))
    {
	if (proxy_query)
	{
#define ISBYTE(v) (((v) & ~0xFF) == 0)
	    
	    if (!ISBYTE(la1) ||
		!ISBYTE(la2) ||
		!ISBYTE(la3) ||
		!ISBYTE(la4) ||
		!ISBYTE(ra1) ||
		!ISBYTE(ra2) ||
		!ISBYTE(ra3) ||
		!ISBYTE(ra4))
	    {
		send_error(fb, cp, buf, "X-INVALID-ADDR");
		return 0;
	    }
	}
	
	if (local_port < 1 || local_port > 65535 ||
	    remote_port < 1 || remote_port > 65535)
	{
	    send_error(fb, cp, buf, "INVALID-PORT");
	    return 0;
	}

	if (reject_pidentd_port && local_port == listen_port)
	{
	    send_error(fb, cp, buf, "HIDDEN-USER");
	    return 0;
	}
	
	kp = kernel_alloc();

	kp->local = *local_addr;
	kp->remote = *remote_addr;

#if 0 /* XXX: FIXME! This is not Casper-IPv6-ok! (call str2addr?) */
	if (proxy_query)
	{
	    if (allow_proxy_query == 1 && 
		kp->local.sin_addr.s_addr != htonl(INADDR_LOOPBACK))
	    {
		kernel_free(kp);
		send_error(fb, cp, buf, "X-INVALID-PROXY-REQUESTER");
		return -1;
	    }

	    kp->local.sin_addr.s_addr = (la1<<24)+(la2<<16)+(la3<<8)+la4;
	    kp->remote.sin_addr.s_addr = (ra1<<24)+(ra2<<16)+(ra3<<8)+ra4;
	}
#endif	
	SGPORT(kp->remote) = htons(remote_port);
	SGPORT(kp->local) = htons(local_port);

	if (enable_socks)
	{
	    char ubuf[256], *user;


	    user = socks_lookup(kp, ubuf, sizeof(ubuf));
	    if (user)
	    {
		if (result_syslog_level > -1)
		    syslog(result_syslog_level,
			   "reply to %s: %s : USERID : UNKNOWN :%s\n",
			   s_inet_ntox(&cp->rsin, nab, sizeof(nab)),
			   buf, user);

		fd_printf(fb, "%s : USERID : UNKNOWN :%s\n",
			  buf, user);
		kernel_free(kp);
		return 0;
	    }
	}
	
	kernel_query(kp);

	if (debug)
	{
	    fprintf(stderr, "kernel_query, status = %d\n", kp->status);
	    if (kp->status == 0)
	    {
		if (kp->ruid != NO_UID)
		    fprintf(stderr, "\truid = %ld\n", (long) kp->ruid);
		if (kp->euid != NO_UID)
		    fprintf(stderr, "\teuid = %ld\n", (long) kp->euid);
		if (kp->pid != NO_PID)
		    fprintf(stderr, "\tpid  = %ld\n", (long) kp->pid);
		if (kp->cmd)
		    fprintf(stderr, "\tcmd  = %s\n", kp->cmd);
		if (kp->argv)
		    fprintf(stderr, "\targv = %s\n", kp->argv);
	    }
	}
	
	if (kp->status == 0)
	    send_result(fb, cp, buf, kp);
	else
	    send_error(fb, cp, buf, "NO-USER");

	kernel_free(kp);
	return 0;
    }

    if (extensions_enabled)
    {
	cmd = s_strtok_r(buf, " ", &bp);
	if (cmd == NULL)
	    goto Fail;
	
	if (strcasecmp("VERSION", cmd) == 0)
	{
	    send_version(fb, cp);
	    return 0;
	}
	
	if (strcasecmp("QUIT", cmd) == 0)
	    return -1;
    }

  Fail:
    send_error(fb, cp, buf, "UNKNOWN-ERROR");
    return -1;
}


static void
timeout_handler(void *arg)
{
    CLIENT *cp = (CLIENT *) arg;

    if (debug)
	fprintf(stderr, "timeout_handler(%08lx)\n", (unsigned long) arg);

    if (cp != NULL)
	shutdown(cp->fd, 2);
}



void
request_handler(CLIENT *cp)
{
    TIMEOUT *tp = NULL;
    FDBUF *fb = NULL;
    char buf[256];


    if (debug)
	fprintf(stderr, "request_handler: fd#%d: start\n", cp->fd);

    if (request_timeout > 0)
	tp = timeout_create(request_timeout, timeout_handler, cp);

    fb = fd_create(cp->fd, FDF_CRLF|FDF_LINEBUF);
    
    while (fd_gets(fb, buf, sizeof(buf)) == 0)
    {
	if (handle_one_request(cp, buf, fb) != 0)
	    break;

	if (!multiquery_enabled)
	    break;
	
	if (tp)
	    timeout_reset(tp, request_timeout);
    }

    timeout_destroy(tp);
    fd_destroy(fb);
    
    if (debug)
	fprintf(stderr, "request_thread: fd#%d: terminating\n", cp->fd);
}
